@@ -1,42 +1,36 @@
-//Copyright (C) 2000 VMELinux Project and Michael Wyrick
-//
-//This program is free software; you can redistribute it and/or
-//modify it under the terms of the GNU General Public License
-//as published by the Free Software Foundation; either version 2
-//of the License, or (at your option) any later version.
-//
-//This program is distributed in the hope that it will be useful,
-//but WITHOUT ANY WARRANTY; without even the implied warranty of
-//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-//GNU General Public License for more details.
-//
-//You should have received a copy of the GNU General Public License
-//along with this program; if not, write to the Free Software
-//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
-//
-//-------------------------------------------------------------------------
+//------------------------------------------------------------------------------  
 //title: Tundra Universe PCI-VME Kernel Driver
-//$date: $
-//-------------------------------------------------------------------------
+//version: Linux 1.1
+//date: March 1999                                                                
+//designer: Michael Wyrick                                                      
+//programmer: Michael Wyrick                                                    
+//platform: Linux 2.4.x
+//language: GCC 2.95 and 3.0
+//module: ca91c042
+//------------------------------------------------------------------------------  
 //  Purpose: Provide a Kernel Driver to Linux for the Universe I and II 
 //           Universe model number ca91c042
 //  Docs:                                  
-//    This driver supports both the Universe and Universe II chips
-//--------------------------------------------------------------------------  
-// $Id: ca91c042.c,v 1.19 1999/11/09 14:47:39 wyrick Exp $
+//    This driver supports both the Universe and Universe II chips                                     
+//------------------------------------------------------------------------------  
+// RCS:
+// $Id: ca91c042.c,v 1.6 2001/10/16 15:16:53 wyrick Exp $
 // $Log: ca91c042.c,v $
+// Revision 1.6  2001/10/16 15:16:53  wyrick
+// Minor Cleanup of Comments
 //
-//---------------------------------------------------------------------------
+//
+//-----------------------------------------------------------------------------
 #define MODULE
 
-static char Version[] = "1.0 99Nov9 Beta";
+static char Version[] = "1.1 2001Oct16";
 
 #include <linux/config.h>
 #include <linux/version.h>
 
 #ifdef CONFIG_MODVERSIONS
-#define MODVERSIONS
-#include <linux/modversions.h>
+  #define MODVERSIONS
+  #include <linux/modversions.h>
 #endif
 
 #include <linux/module.h>
@@ -50,11 +44,6 @@ static char Version[] = "1.0 99Nov9 Beta";
 
 #include "ca91c042.h"
 
-#if LINUX_VERSION_CODE < 0x020100
-#define	ioremap	vremap
-#define	iounmap	vfree
-#endif
-
 //----------------------------------------------------------------------------
 // Prototypes
 //----------------------------------------------------------------------------
@@ -76,16 +65,13 @@ static struct proc_dir_entry *uni_procdir;
 
 static struct file_operations uni_fops = 
 {
-  uni_llseek,
-  uni_read,
-  uni_write,
-  NULL,         // uni_readdir
-  uni_poll,     // uni_poll   
-  uni_ioctl,
-  NULL,         // uni_mmap     
-  uni_open,
-  NULL,         // Flush
-  uni_release 
+  llseek:   uni_llseek,
+  read:     uni_read,
+  write:    uni_write,
+  poll:     uni_poll,     // uni_poll   
+  ioctl:    uni_ioctl,
+  open:     uni_open,
+  release:  uni_release 
 };
 
 static int aCTL[] = {LSI0_CTL, LSI1_CTL, LSI2_CTL, LSI3_CTL};
@@ -96,7 +82,7 @@ static int aTO[]  = {LSI0_TO,  LSI1_TO,  LSI2_TO,  LSI3_TO};
 //----------------------------------------------------------------------------
 // Vars and Defines
 //----------------------------------------------------------------------------
-#define UNI_MAJOR   	  70
+#define UNI_MAJOR     221
 #define MAX_MINOR 	    8
 #define CONTROL_MINOR   8
 
@@ -104,17 +90,21 @@ static int aTO[]  = {LSI0_TO,  LSI1_TO,  LSI2_TO,  LSI3_TO};
 #define MODE_PROGRAMMED 1
 #define MODE_DMA 	      2
 
+#define DMATYPE_SINGLE  1
+#define DMATYPE_LLIST   2
+
 static int OkToWrite[MAX_MINOR + 1];        // Can I write to the Hardware
 
 static int opened[MAX_MINOR + 1];
-static int mode[MAX_MINOR + 1];		          // DMA or Programmed I/O
+static int mode[MAX_MINOR + 1];             // DMA or Programmed I/O
 
 static unsigned long DMA[MAX_MINOR + 1];    // DMA Control Reg
 
 static int status;
+static int DMAType = 0;
 static int irq = 0;
 static char *baseaddr  = 0;
-static char *image_ba[MAX_MINOR+1];	        // Base PCI Address
+static char *image_ba[MAX_MINOR+1];         // Base PCI Address
 
 static unsigned int image_ptr[MAX_MINOR+1];
 static unsigned int image_va[MAX_MINOR+1];
@@ -125,7 +115,7 @@ static short vmeirqbit[7]     = {IRQ_VIRQ1, IRQ_VIRQ2, IRQ_VIRQ3, IRQ_VIRQ4,
                                  IRQ_VIRQ5, IRQ_VIRQ6, IRQ_VIRQ7};               
 static int vmevec[7]          = {V1_STATID, V2_STATID, V3_STATID, V4_STATID,
                                  V5_STATID, V6_STATID, V7_STATID};
-			      
+
 // Status Vars
 static unsigned int reads  = 0;
 static unsigned int writes = 0;
@@ -144,7 +134,7 @@ static int uni_procinfo(char *buf, char **start, off_t fpos, int lenght, int *eo
 {
   char *p;
   unsigned int temp1,temp2,x;
-  
+
   p = buf;
   p += sprintf(p,"Universe driver %s\n",Version);
 
@@ -152,7 +142,7 @@ static int uni_procinfo(char *buf, char **start, off_t fpos, int lenght, int *eo
   p += sprintf(p,"  Stats  reads = %i  writes = %i  ioctls = %i\n",
                  reads,writes,ioctls);
 
-  for(x=0;x<4;x+=2) {
+  for (x=0;x<4;x+=2) {
     temp1 = readl(baseaddr+aCTL[x]);
     temp2 = readl(baseaddr+aCTL[x+1]);
     p += sprintf(p,"  LSI%i_CTL = %08X    LSI%i_CTL = %08X\n",x,temp1,x+1,temp2);
@@ -166,27 +156,27 @@ static int uni_procinfo(char *buf, char **start, off_t fpos, int lenght, int *eo
     temp2 = readl(baseaddr+aTO[x+1]);
     p += sprintf(p,"  LSI%i_TO  = %08X    LSI%i_TO  = %08X\n",x,temp1,x+1,temp2);
   }  
-  
-  for(x=0;x<3;x+=2)
+
+  for (x=0;x<3;x+=2)
     p += sprintf(p,"  image_va%i   = %08X     image_va%i   = %08X\n",
-      x,image_va[x],x+1,image_va[x+1]); 
+                 x,image_va[x],x+1,image_va[x+1]); 
   p += sprintf(p,"\nDriver Program Status:\n");  
-      
-  for(x=0;x<3;x+=2)
+
+  for (x=0;x<3;x+=2)
     p += sprintf(p,"  DMACTL %i    = %08lX DMACTL %i    = %08lX\n",
-      x,DMA[x],x+1,DMA[x+1]); 	     
-  for(x=0;x<3;x+=2)
+                 x,DMA[x],x+1,DMA[x+1]);        
+  for (x=0;x<3;x+=2)
     p += sprintf(p,"  OkToWrite %i = %1X        OkToWrite %i = %1X\n",
-      x,OkToWrite[x],x+1,OkToWrite[x+1]); 
-  for(x=0;x<3;x+=2)
+                 x,OkToWrite[x],x+1,OkToWrite[x+1]); 
+  for (x=0;x<3;x+=2)
     p += sprintf(p,"  Mode %i      = %1X        Mode %i      = %1X\n",
-      x,mode[x],x+1,mode[x+1]); 
+                 x,mode[x],x+1,mode[x+1]); 
 
   p += sprintf(p,"\n");  
 
   temp1 = 0;
   p += sprintf(p, "VMEIrqs Assigned: ");
-  for(x=0;x<7;x++) {
+  for (x=0;x<7;x++) {
     if (vmeirqs[x] != NULL) {
       p += sprintf(p, "%i ",x+1);
       temp1++;
@@ -194,9 +184,9 @@ static int uni_procinfo(char *buf, char **start, off_t fpos, int lenght, int *eo
   }
   if (temp1 == 0)
     p += sprintf(p, "none\n");
-  else  
+  else
     p += sprintf(p,"\n");  
-  
+
   *eof = 1;
   return p - buf;
 }
@@ -217,7 +207,7 @@ static void unregister_proc()
 {
   remove_proc_entry("ca91c042",0);
 }
-	     					   
+
 //----------------------------------------------------------------------------
 //
 //  uni_poll()
@@ -236,7 +226,7 @@ static unsigned int uni_poll(struct file* file, poll_table* wait)
 static int uni_open(struct inode *inode,struct file *file)
 {
   unsigned int minor = MINOR(inode->i_rdev);
-  
+
   if (minor > MAX_MINOR) {
     return(-ENODEV);
   }
@@ -244,12 +234,12 @@ static int uni_open(struct inode *inode,struct file *file)
   if (minor == CONTROL_MINOR) {
     opened[minor]++;
     return(0);
-  }  
-  
+  }
+
   if (!opened[minor]) {
     opened[minor] = 1;
     return(0);
-  } else 
+  } else
     return(-EBUSY);
 }
 
@@ -261,9 +251,9 @@ static int uni_open(struct inode *inode,struct file *file)
 static int uni_release(struct inode *inode,struct file *file)
 {
   unsigned int minor = MINOR(inode->i_rdev);
-  
+
   opened[minor]--;
-  
+
   return 0;
 }
 
@@ -276,8 +266,8 @@ static long long uni_llseek(struct file *file,loff_t offset,int whence)
 {
   unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
   unsigned int toffset,base,paddr;
-  
-  if (whence == SEEK_SET) {	       
+
+  if (whence == SEEK_SET) {
     if (minor == CONTROL_MINOR) {
       image_ptr[minor] = offset;
     } else {
@@ -306,46 +296,46 @@ static ssize_t uni_read(struct file *file, char *buf, size_t count, loff_t *ppos
   char *temp = buf;
   unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);  
   int p; 
-  
+
   unsigned char  vc;
   unsigned short vs;
   unsigned int   vl;
-  
+
   char *DMA_Buffer;
   unsigned int DMA_Buffer_Size = 0, 
-               order = 0,
-               a_size = 0,
-               dma_align = 0,
-               timeout = 0;						
-  int val,
-      pci = 0,
-      vme = 0,
-      okcount = 0;
-  
+    order     = 0,
+    a_size    = 0,
+    dma_align = 0,
+    timeout   = 0;                     
+  int val     = 0,
+    pci       = 0,
+    vme       = 0,
+    okcount   = 0;
+
   if (minor == CONTROL_MINOR) {
     p = (int)image_ptr[minor];
     v = readl(baseaddr+p);    
     __copy_to_user(temp,&v,4);
-  } else { 
-    reads++;  	       	    
+  } else {
+    reads++;                
     if (OkToWrite[minor]) {
       if (mode[minor] == MODE_PROGRAMMED) {
         numt = count;
         remain = count;
-      
+
         // Calc the number of longs we need
         numt = count / 4;
         remain = count % 4;
         for (x=0;x<numt;x++) {
           vl = readl(image_ptr[minor]);    
-	  
+
           // Lets Check for a Bus Error
           tmp = readl(baseaddr+PCI_CSR);
           if (tmp & 0x08000000) {
             writel(tmp,baseaddr+PCI_CSR);
-	    return(okcount);
+            return(okcount);
           } else
-	    okcount += 4;
+            okcount += 4;
 
           __copy_to_user(temp,&vl,4);
           image_ptr[minor]+=4;
@@ -362,15 +352,15 @@ static ssize_t uni_read(struct file *file, char *buf, size_t count, loff_t *ppos
           tmp = readl(baseaddr+PCI_CSR);
           if (tmp & 0x08000000) {
             writel(tmp,baseaddr+PCI_CSR);
-	    return(okcount);
+            return(okcount);
           } else
-	    okcount += 2;
+            okcount += 2;
 
           __copy_to_user(temp,&vs,2);
           image_ptr[minor]+=2;
           temp+=2;
         }  
-        
+
         for (x=0;x<remain;x++) {
           vc = readb(image_ptr[minor]);    
 
@@ -378,67 +368,67 @@ static ssize_t uni_read(struct file *file, char *buf, size_t count, loff_t *ppos
           tmp = readl(baseaddr+PCI_CSR);
           if (tmp & 0x08000000) {
             writel(tmp,baseaddr+PCI_CSR);
-	    return(okcount);
+            return(okcount);
           } else
-	    okcount++;
+            okcount++;
 
           __copy_to_user(temp,&vc,1);
           image_ptr[minor]+=1;
           temp+=1;
         }  
 
-      } else if (mode[minor] == MODE_DMA) {   
-      // ------------------------------------------------------------------
-      //
-      // ------------------------------------------------------------------      
+      } else if (mode[minor] == MODE_DMA) {
+        // ------------------------------------------------------------------
+        //
+        // ------------------------------------------------------------------      
         // Wait for DMA to finish, This needs to be changed
         val = readl(baseaddr+DGCS);
-        while ((val & 0x00008000) && (timeout++ < 1000000)) 
+        while ((val & 0x00008000) && (timeout++ < 1000000))
           val = readl(baseaddr+DGCS);
 
         // VME Address						      
-	vme = image_va[minor] + (image_ptr[minor] - (unsigned int)image_ba[minor]);  
-	dma_align = vme % 8;  
-	  
+        vme = image_va[minor] + (image_ptr[minor] - (unsigned int)image_ba[minor]);  
+        dma_align = vme % 8;  
+
         // Setup DMA Buffer to read data into
-	DMA_Buffer_Size = count + ((int)image_ptr % 8) + dma_align;         
-	a_size = PAGE_SIZE;
-	while (a_size < DMA_Buffer_Size) {
-	  order++;
-	  a_size <<= 1;
-	}
-	DMA_Buffer = (char *)__get_dma_pages(GFP_KERNEL,order);	 
-	
-	// PCI Address
-	pci = virt_to_bus(DMA_Buffer) + dma_align;
-	
-	// Setup DMA regs
-	writel(DMA[minor],baseaddr+DCTL);		// Setup Control Reg
-	writel(count,baseaddr+DTBC);	     		// Count	    
-	writel(pci,baseaddr+DLA);	 		// PCI Address
-	writel(vme,baseaddr+DVA);		 	// VME Address
-	
-	// Start DMA
-	writel(0x80006F00,baseaddr+DGCS);		// GO
-	
+        DMA_Buffer_Size = count + ((int)image_ptr % 8) + dma_align;         
+        a_size = PAGE_SIZE;
+        while (a_size < DMA_Buffer_Size) {
+          order++;
+          a_size <<= 1;
+        }
+        DMA_Buffer = (char *)__get_dma_pages(GFP_KERNEL,order);  
+
+        // PCI Address
+        pci = virt_to_bus(DMA_Buffer) + dma_align;
+
+        // Setup DMA regs
+        writel(DMA[minor],baseaddr+DCTL);       // Setup Control Reg
+        writel(count,baseaddr+DTBC);            // Count	    
+        writel(pci,baseaddr+DLA);               // PCI Address
+        writel(vme,baseaddr+DVA);               // VME Address
+
+        // Start DMA
+        writel(0x80006F00,baseaddr+DGCS);       // GO
+
         // Wait for DMA to finish, This needs to be changed
         val = readl(baseaddr+DGCS);
-        while ((val & 0x00008000) && (timeout++ < 100000)) 
+        while ((val & 0x00008000) && (timeout++ < 100000))
           val = readl(baseaddr+DGCS);
-	
-	if (timeout == 100000)
-	  printk("<<ca91c042 DMA Timed out>>\n");  
-	  
-	if (! (val & 0x00000800)) {  // An Error Happened
-	  count -= readl(baseaddr+DTBC);	
-	}
-		
+
+        if (timeout == 100000)
+          printk("<<ca91c042 DMA Timed out>>\n");
+
+        if (! (val & 0x00000800)) {  // An Error Happened
+          count -= readl(baseaddr+DTBC);    
+        }
+
         image_ptr[minor] += count;
-        		
-   	// Copy pages to User Memory
-	__copy_to_user(temp,DMA_Buffer+dma_align,count);
-	
-	free_pages((unsigned long)DMA_Buffer,order);
+
+        // Copy pages to User Memory
+        __copy_to_user(temp,DMA_Buffer+dma_align,count);
+
+        free_pages((unsigned long)DMA_Buffer,order);
       }           // end of MODE_DMA
     }             // end of OKtoWrite
   }  
@@ -457,34 +447,34 @@ static ssize_t uni_write(struct file *file, const char *buf, size_t count, loff_
   unsigned int numt,remain,tmp;
   char *temp = (char *)buf;
   unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
-  
+
   unsigned char  vc;
   unsigned short vs;
   unsigned int   vl;
-  
+
   char *DMA_Buffer;
   unsigned int DMA_Buffer_Size = 0, 
-               order = 0,
-               a_size = 0,
-               dma_align = 0,
-               timeout = 0;						
+    order     = 0,
+    a_size    = 0,
+    dma_align = 0,
+    timeout   = 0;                     
   int val,
-      pci = 0,
-      vme = 0,
-      okcount = 0;
-  
+    pci       = 0,
+    vme       = 0,
+    okcount   = 0;
+
   writes++;
   if (minor == CONTROL_MINOR) {
     __copy_from_user(&vl,temp,4);
     p = (int)image_ptr[minor];
     writel(vl,baseaddr+p);
-  } else { 
+  } else {
     if (OkToWrite[minor]) {
       if (mode[minor] == MODE_PROGRAMMED) {
         // Calc the number of longs we need
         numt = count;
         remain = count;
-	
+
         numt = count / 4;
         remain = count % 4;
         for (x=0;x<numt;x++) {
@@ -495,9 +485,9 @@ static ssize_t uni_write(struct file *file, const char *buf, size_t count, loff_
           tmp = readl(baseaddr+PCI_CSR);
           if (tmp & 0x08000000) {
             writel(tmp,baseaddr+PCI_CSR);
-	    return(okcount);
+            return(okcount);
           } else
-	    okcount += 4;
+            okcount += 4;
 
           image_ptr[minor]+=4;
           temp+=4;
@@ -506,39 +496,39 @@ static ssize_t uni_write(struct file *file, const char *buf, size_t count, loff_
         // Calc the number of Words we need
         numt = remain / 2;
         remain = remain % 2;
-    
+
         for (x=0;x<numt;x++) {
           __copy_from_user(&vs,temp,2);
           writew(vs,image_ptr[minor]);
-	  
+
           // Lets Check for a Bus Error
           tmp = readl(baseaddr+PCI_CSR);
           if (tmp & 0x08000000) {
             writel(tmp,baseaddr+PCI_CSR);
-	    return(okcount);
+            return(okcount);
           } else
-	    okcount += 2;
-	    
+            okcount += 2;
+
           image_ptr[minor]+=2;
           temp+=2;
         }  
-      
+
         for (x=0;x<remain;x++) {
           __copy_from_user(&vc,temp,1);
           writeb(vc,image_ptr[minor]);
-	  
+
           // Lets Check for a Bus Error
           tmp = readl(baseaddr+PCI_CSR);
           if (tmp & 0x08000000) {
             writel(tmp,baseaddr+PCI_CSR);
-	    return(okcount);
+            return(okcount);
           } else
-	    okcount += 2;
-	    
+            okcount += 2;
+
           image_ptr[minor]+=1;
           temp+=1;
         }  
-	
+
         // Lets Check for a Bus Error
         tmp = readl(baseaddr+PCI_CSR);
         if (tmp & 0x08000000) {  // S_TA is Set
@@ -546,60 +536,59 @@ static ssize_t uni_write(struct file *file, const char *buf, size_t count, loff_
           count = 0;
         }
 
-      }	else if (mode[minor] == MODE_DMA) {
-      // ------------------------------------------------------------------
-      //
-      // ------------------------------------------------------------------      
+      } else if (mode[minor] == MODE_DMA) {
+        // ------------------------------------------------------------------
+        //
+        // ------------------------------------------------------------------      
         // Wait for DMA to finish, This needs to be changed
         val = readl(baseaddr+DGCS);
-        while ((val & 0x00008000) && (timeout++ < 100000)) 
+        while ((val & 0x00008000) && (timeout++ < 100000))
           val = readl(baseaddr+DGCS);
-	  
+
         // Setup DMA Buffer to write data into
         // VME Address						      
-	vme = image_va[minor] + (image_ptr[minor] - (unsigned int)image_ba[minor]);  
-	dma_align = vme % 8;  
+        vme = image_va[minor] + (image_ptr[minor] - (unsigned int)image_ba[minor]);  
+        dma_align = vme % 8;  
+
+        DMA_Buffer_Size = count + ((int)image_ptr % 8) + dma_align;         
+        a_size = PAGE_SIZE;
+        while (a_size < DMA_Buffer_Size) {
+          order++;
+          a_size <<= 1;
+        }
+        DMA_Buffer = (char *)__get_dma_pages(GFP_KERNEL,order);  
+
+        // Copy User Memory into buffer
+        __copy_from_user(DMA_Buffer + dma_align,temp,count);
+
+        // PCI Address
+        pci = virt_to_bus(DMA_Buffer) + dma_align;
+
+        // Setup DMA regs
+        writel(DMA[minor]|0x80000000,baseaddr+DCTL);  // Setup Control Reg
+        writel(count,baseaddr+DTBC);                  // Count	    
+        writel(pci,baseaddr+DLA);                     // PCI Address
+        writel(vme,baseaddr+DVA);                     // VME Address
+
+        // Start DMA
+        writel(0x80006F00,baseaddr+DGCS);             // GO
 
-	DMA_Buffer_Size = count + ((int)image_ptr % 8) + dma_align;         
-	a_size = PAGE_SIZE;
-	while (a_size < DMA_Buffer_Size) {
-	  order++;
-	  a_size <<= 1;
-	}
-	DMA_Buffer = (char *)__get_dma_pages(GFP_KERNEL,order);	 
-	
-   	// Copy User Memory into buffer
-	__copy_from_user(DMA_Buffer + dma_align,temp,count);
-	
-	
-	// PCI Address
-	pci = virt_to_bus(DMA_Buffer) + dma_align;
-	
-	// Setup DMA regs
-	writel(DMA[minor]|0x80000000,baseaddr+DCTL);	// Setup Control Reg
-	writel(count,baseaddr+DTBC);	     		// Count	    
-	writel(pci,baseaddr+DLA); 			// PCI Address
-	writel(vme,baseaddr+DVA);		 	// VME Address
-	
-	// Start DMA
-	writel(0x80006F00,baseaddr+DGCS);		// GO
-	
         // Wait for DMA to finish, This needs to be changed
         val = readl(baseaddr+DGCS);
-        while ((val & 0x00008000) && (timeout++ < 100000)) 
+        while ((val & 0x00008000) && (timeout++ < 100000))
           val = readl(baseaddr+DGCS);
 
-	if (timeout == 100000)
-	  printk("<<ca91c042 DMA Timed out>>\n");  
-	  
-	if (! (val & 0x00000800)) {  // An Error Happened
+        if (timeout == 100000)
+          printk("<<ca91c042 DMA Timed out>>\n");
+
+        if (! (val & 0x00000800)) {  // An Error Happened
           // The Universe Seems to return an invalid value in DTBC on
           // Bus Errors during DMA, so invalidate the count
           count = 0;
-	}
+        }
         image_ptr[minor] += count;
-	
-	free_pages((unsigned long)DMA_Buffer,order);
+
+        free_pages((unsigned long)DMA_Buffer,order);
       }           // end of MODE_DMA
     }  //OKtoWrite
   }  
@@ -614,66 +603,64 @@ static int uni_ioctl(struct inode *inode,struct file *file,unsigned int cmd, uns
 {
   unsigned int minor = MINOR(inode->i_rdev);
   unsigned int sizetomap = 0, to = 0, bs = 0;
-  
+
   ioctls++;
   switch (cmd) {
-    case IOCTL_SET_CTL:
-//      if (minor != 3)
-//         arg &= 0xFFFFEFFF;
-      writel(arg,baseaddr+aCTL[minor]);
-      
-      // Lets compute and save the DMA CTL Register
-      DMA[minor] = arg & 0x00FFFF00;
-      break;
-      	    
-    case IOCTL_SET_MODE:
-      mode[minor] = arg;
-      break;	    
-    
+  case IOCTL_SET_CTL:
+    writel(arg,baseaddr+aCTL[minor]);
+
+    // Lets compute and save the DMA CTL Register
+    DMA[minor] = arg & 0x00FFFF00;
+    break;
+
+  case IOCTL_SET_MODE:
+    mode[minor] = arg;
+    break;        
+
     // Lets Map the VME Bus to the PCI Bus
     //
     // << NOTE >> BD Must already be set before you call this!!!!
     //
     //
-    case IOCTL_SET_BS:
-      writel(arg,baseaddr+aBS[minor]);
-      if (image_ba[minor])
-        iounmap(image_ba[minor]);
-  
-      // This uses the BD Register to Find the size of the Image Mapping		
-      sizetomap = readl(baseaddr+aBD[minor]);
-      sizetomap -= arg;
-      		
-      image_ba[minor] = (char *)ioremap(arg,sizetomap);
-      if (!image_ba[minor]) {
-        OkToWrite[minor] = 0;     
-        return -1;
-      }
-      image_ptr[minor] = (int)image_ba[minor];
-      OkToWrite[minor] = 1;
-      
-      // Calculate the VME Address
-      bs = readl(baseaddr+aBS[minor]);
-      to = readl(baseaddr+aTO[minor]);
-      image_va[minor]  = bs + to;
-      break;
-      
-    case IOCTL_SET_BD:
-      writel(arg,baseaddr+aBD[minor]);
-      break;  
-    case IOCTL_SET_TO:
-      writel(arg,baseaddr+aTO[minor]);
-      				      
-      // Calculate the VME Address
-      bs = readl(baseaddr+aBS[minor]);
-      to = readl(baseaddr+aTO[minor]);
-      image_va[minor]  = bs + to;
-      break;  
-    default:
-      if (cmd < 0x1000) {  // This is a Register value so write to it.
-        writel(arg,baseaddr+cmd);
-      }  
-      break;  
+  case IOCTL_SET_BS:
+    writel(arg,baseaddr+aBS[minor]);
+    if (image_ba[minor])
+      iounmap(image_ba[minor]);
+
+    // This uses the BD Register to Find the size of the Image Mapping		
+    sizetomap = readl(baseaddr+aBD[minor]);
+    sizetomap -= arg;
+
+    image_ba[minor] = (char *)ioremap(arg,sizetomap);
+    if (!image_ba[minor]) {
+      OkToWrite[minor] = 0;     
+      return -1;
+    }
+    image_ptr[minor] = (int)image_ba[minor];
+    OkToWrite[minor] = 1;
+
+    // Calculate the VME Address
+    bs = readl(baseaddr+aBS[minor]);
+    to = readl(baseaddr+aTO[minor]);
+    image_va[minor]  = bs + to;
+    break;
+
+  case IOCTL_SET_BD:
+    writel(arg,baseaddr+aBD[minor]);
+    break;  
+  case IOCTL_SET_TO:
+    writel(arg,baseaddr+aTO[minor]);
+
+    // Calculate the VME Address
+    bs = readl(baseaddr+aBS[minor]);
+    to = readl(baseaddr+aTO[minor]);
+    image_va[minor]  = bs + to;
+    break;  
+  default:
+    if (cmd < 0x1000) {  // This is a Register value so write to it.
+      writel(arg,baseaddr+cmd);
+    }
+    break;  
   }
   return(0);
 }
@@ -693,7 +680,7 @@ void PrintCmdPacketList(TDMA_Cmd_Packet* cpl)
   int x = 0;
   int done = 0;
 
-  while(!done && (x < 10)) {
+  while (!done && (x < 10)) {
     x++;
     sprintf(buff,"<ca91c042> (%i) dctl=%08X, dtbc=%08X\n",x,p->dctl,p->dtbc);
     printk(buff);
@@ -703,7 +690,7 @@ void PrintCmdPacketList(TDMA_Cmd_Packet* cpl)
     printk(buff);
     done = (p->dcpp & 0x00000001);
     p = bus_to_virt((int)(TDMA_Cmd_Packet*)(p->dcpp & 0xFFFFFFF8));  // Next in Line Please
-	}
+  }
 }
 
 //-----------------------------------------------------------------------------
@@ -761,14 +748,16 @@ void DMA_irq_handler(void)
     sprintf(buf,"<<ca91c042>> DVA=%08X\n",val);  
     printk(buf);
 
-    printk("<<ca91c042>> CmdPacketList sent to DMARead\n");
-    PrintCmdPacketList(debugptr);
-    
-    printk("<<ca91c042>> CmdPacketList as seen by DCPP\n");
-    cpl = (TDMA_Cmd_Packet*)readl(baseaddr+DCPP);   // Command Packet Pointer
-    cpl = (TDMA_Cmd_Packet*)((int)cpl & ~0x03);
+    if (DMAType == DMATYPE_LLIST) {
+      printk("<<ca91c042>> CmdPacketList sent to DMARead\n");
+      PrintCmdPacketList(debugptr);
 
-    PrintCmdPacketList(cpl);
+      printk("<<ca91c042>> CmdPacketList as seen by DCPP\n");
+      cpl = (TDMA_Cmd_Packet*)readl(baseaddr+DCPP);   // Command Packet Pointer
+      cpl = (TDMA_Cmd_Packet*)((int)cpl & ~0x03);
+
+      PrintCmdPacketList(cpl);
+    }
     cbv = 1;  // Call Back value is set to 1 to show error condition
   }
 
@@ -798,16 +787,17 @@ void LERR_irq_handler(void)
   val = readl(baseaddr+DGCS);
 
   if (!(val & 0x00000800)) {
-    sprintf(buf,"<<ca91c042>> DMA Read Error DGCS=%08X\n",val);  
+    sprintf(buf,"<<ca91c042>> LERR_irq_handler DMA Read Error DGCS=%08X\n",val);  
     printk(buf);
 
-    cpl = (TDMA_Cmd_Packet*)readl(baseaddr+DCPP);   // Command Packet Pointer
-    cpl = (TDMA_Cmd_Packet*)((int)cpl & ~0x03);
+    if (DMAType == DMATYPE_LLIST) {
+      cpl = (TDMA_Cmd_Packet*)readl(baseaddr+DCPP);   // Command Packet Pointer
+      cpl = (TDMA_Cmd_Packet*)((int)cpl & ~0x03);
+      PrintCmdPacketList(cpl);
 
-    PrintCmdPacketList(cpl);
-
-    sprintf(buf,"<<ca91c042>> DMAReadCallBack Request of cmdpack=0x%08X\n",cpl);
-    printk(buf);
+      sprintf(buf,"<<ca91c042>> DMAReadCallBack Request of cmdpack=0x%08X\n",(int)cpl);
+      printk(buf);
+    }
   }
 
   if (DMACallBackFunc)
@@ -834,16 +824,18 @@ void VERR_irq_handler(void)
   val = readl(baseaddr+DGCS);
 
   if (!(val & 0x00000800)) {
-    sprintf(buf,"<<ca91c042>> DMA Read Error DGCS=%08X\n",val);  
+    sprintf(buf,"<<ca91c042>> VERR_irq_handler DMA Read Error DGCS=%08X\n",val);  
     printk(buf);
 
-    cpl = (TDMA_Cmd_Packet*)readl(baseaddr+DCPP);   // Command Packet Pointer
-    cpl = (TDMA_Cmd_Packet*)((int)cpl & ~0x03);
+    if (DMAType == DMATYPE_LLIST) {
+      cpl = (TDMA_Cmd_Packet*)readl(baseaddr+DCPP);   // Command Packet Pointer
+      cpl = (TDMA_Cmd_Packet*)((int)cpl & ~0x03);
 
-    PrintCmdPacketList(cpl);
+      PrintCmdPacketList(cpl);
 
-    sprintf(buf,"<<ca91c042>> DMAReadCallBack Request of cmdpack=0x%08X\n",cpl);
-    printk(buf);
+      sprintf(buf,"<<ca91c042>> DMAReadCallBack Request of cmdpack=0x%08X\n",(int)cpl);
+      printk(buf);
+    }
   }
 
   if (DMACallBackFunc)
@@ -873,13 +865,13 @@ static void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
   if (stat & 0x0400)
     VERR_irq_handler();
 
-  for(i=0;i<7;i++) {
+  for (i=0;i<7;i++) {
     vector = readl(baseaddr+vmevec[i]);
     if (stat & vmeirqbit[i] & enable) {
       if (vmeirqs[i] != NULL) {
         vmeirqs[i](i+1, vector, dev_id, regs);
       }
-    }  
+    }
   }
 
   writel(stat, baseaddr+LINT_STAT);                // Clear all pending ints
@@ -894,18 +886,18 @@ static void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
 // History    : 
 //-----------------------------------------------------------------------------
 void cleanup_module(void)
-{	       
+{          
   int x;
   int pcivector;
 
-  writel(0,baseaddr+LINT_EN); 			        // Turn off Ints
+  writel(0,baseaddr+LINT_EN);                   // Turn off Ints
   pcivector = readl(baseaddr+PCI_MISC1) & 0x000000FF; 
-  free_irq(pcivector,NULL);				// Free Vector
-      
+  free_irq(pcivector,NULL);             // Free Vector
+
   for (x=1;x<MAX_MINOR+1;x++) {
     if (image_ba[x])
       iounmap(image_ba[x]);
-    }    
+  }    
   iounmap(baseaddr);
   unregister_proc();
   unregister_chrdev(UNI_MAJOR, "uni");
@@ -919,31 +911,30 @@ int init_module(void)
   int x, result;
   unsigned int temp, ba;
   char vstr[80];
-  
+
   struct pci_dev *uni_pci_dev = NULL;
 
   sprintf(vstr,"Tundra Universe PCI-VME Bridge Driver %s\n",Version);
   printk(vstr);
-  printk("  Beta Version. Support for Module Interrupts added\n");
-  printk("  Copyright 1997-1999, Michael J. Wyrick\n");
-    
+  printk("  Copyright 1997-2001, Michael J. Wyrick\n");
+
   if ((uni_pci_dev = pci_find_device(PCI_VENDOR_ID_TUNDRA,
                                      PCI_DEVICE_ID_TUNDRA_CA91C042, 
                                      uni_pci_dev))) {
     printk("  Universe device found.");
-      
+
     // Lets turn Latency off
     pci_write_config_dword(uni_pci_dev, PCI_MISC0, 0);
-      
+
     // Display PCI Registers  
     pci_read_config_dword(uni_pci_dev, PCI_CSR, &status);
     printk("  Vendor = %04X  Device = %04X  Status = %08X\n",
-              uni_pci_dev->vendor,uni_pci_dev->device,status);
+           uni_pci_dev->vendor,uni_pci_dev->device,status);
     printk("  Class = %08X\n",uni_pci_dev->class);
-      
+
     pci_read_config_dword(uni_pci_dev, PCI_MISC0, &temp);
     printk("  Misc0 = %08X\n",temp);      
-      
+
     // Setup Universe Config Space
     // This is a 4k wide memory area that need to be mapped into the kernel
     // virtual memory space so we can access it.
@@ -954,7 +945,7 @@ int init_module(void)
       printk("  vremap failed to map Universe to Kernel Space.\r");
       return 1;
     }
-   
+
     // Check to see if the Mapping Worked out
     temp = readl(baseaddr);
     printk("  Read via mapping, PCI_ID = %08X\n",temp);       
@@ -962,17 +953,23 @@ int init_module(void)
       printk("  Universe Chip Failed to Return PCI_ID in Memory Map.\n");
       return 1;
     }
-      
+
     // OK, Every this is ok so lets turn off the windows
     writel(0x00800000,baseaddr+LSI0_CTL);     
     writel(0x00800000,baseaddr+LSI1_CTL);     
     writel(0x00800000,baseaddr+LSI2_CTL);     
     writel(0x00800000,baseaddr+LSI3_CTL);     
-      
+
+    // Write to Misc Register
+    // Set VME Bus Time-out
+    //   Arbitration Mode
+    //   DTACK Enable
+    writel(0x15060000,baseaddr+MISC_CTL);     
+
     // Lets turn off interrupts
     writel(0x00000000,baseaddr+LINT_EN);   // Disable interrupts in the Universe first
     writel(0x0000FFFF,baseaddr+LINT_STAT); // Clear Any Pending Interrupts
-    
+
     pci_read_config_dword(uni_pci_dev, PCI_INTERRUPT_LINE, &irq);
     irq &= 0x000000FF;                    // Only a byte in size
     result = request_irq(irq, irq_handler, SA_INTERRUPT, "VMEBus (ca91c042)", NULL);
@@ -981,13 +978,13 @@ int init_module(void)
     } else {
       writel(0x0000, baseaddr+LINT_MAP0);    // Map all ints to 0
       writel(0x0000, baseaddr+LINT_MAP1);    // Map all ints to 0
-		}
-   
+    }
+
   } else {
     printk("  Universe device not found on PCI Bus.\n");
     return 1;
   }
-     
+
   if (register_chrdev(UNI_MAJOR, "uni", &uni_fops)) {
     printk("  Error getting Major Number for Drivers\n");
     iounmap(baseaddr);
@@ -995,19 +992,20 @@ int init_module(void)
   } else {
     printk("  Tundra Loaded.\n"); 
   }
-  
+
   register_proc();
- 
-  for(x=0;x<MAX_MINOR+1;x++) {
-    image_ba[x]  = 0;	     		// Not defined
-    image_ptr[x] = 0;	     		// Not defined
-    opened[x]    = 0;	     		// Closed
-    OkToWrite[x] = 0;	     		// Not OK
-    mode[x]      = MODE_UNDEFINED;   	// Undefined
+
+  for (x=0;x<MAX_MINOR+1;x++) {
+    image_ba[x]  = 0;               // Not defined
+    image_ptr[x] = 0;               // Not defined
+    opened[x]    = 0;               // Closed
+    OkToWrite[x] = 0;               // Not OK
+    mode[x]      = MODE_UNDEFINED;  // Undefined
   }  
 
   // Setup the DMA Timer
   init_timer(&DMA_timer);
+  DMA_timer.function = DMA_status;
 
   // Enable DMA Interrupts
   writel(0x0700, baseaddr+LINT_EN);
@@ -1017,7 +1015,7 @@ int init_module(void)
 
 //-----------------------------------------------------------------------------
 //
-// Public Functions
+//  Public Interface
 //
 //-----------------------------------------------------------------------------
 //-----------------------------------------------------------------------------
@@ -1091,7 +1089,7 @@ void disable_vmeirq(unsigned int irq)
 //-----------------------------------------------------------------------------
 int request_vmeirq(unsigned int irq, TirqHandler handler)
 {
-  if ((irq > 0) && (irq < 8)) {
+  if ((irq >= 1) && (irq <= 7)) {
     if (vmeirqs[irq-1] == NULL) {
       vmeirqs[irq-1] = handler;
       return 0;
@@ -1112,7 +1110,7 @@ int request_vmeirq(unsigned int irq, TirqHandler handler)
 //-----------------------------------------------------------------------------
 void free_vmeirq(unsigned int irq)
 {
-  if ((irq >=1) && (irq <= 7))
+  if ((irq >= 1) && (irq <= 7))
     vmeirqs[irq-1] = NULL;
 }
 
@@ -1179,29 +1177,22 @@ void unmapvme(char *ptr, int image)
 }
 
 //-----------------------------------------------------------------------------
-// Function   : mapvme
-// Inputs     : unsigned int pci
-//              unsigned int vme 
-//              unsigned int count
-//              int ctl
-// Outputs    : int - Number of Bytes Transfered
+// Function   : VME_DMA
+// Inputs     : unsigned int pci   - PCI Address of Image Map (See howto)
+//              unsigned int vme   - VME Address of Image (See howto)
+//              unsigned int count - Size of Image
+//              int image          - Image Number to Use (See howto)
+//              int ctl            - VME Address Space (See howto)
+// Outputs    : void
 // Description: 
-// Remarks    : Too many to mention right now, PCI is a virtual address
-//              THIS WILL BLOCK UNTIL THE TRANSFER IS COMPLETE
-//              we need to change this to not block at some point
-//              pci and vme MUST be have the same lower 3 bits, this is a 
-//              requirement of the Universe
+// Remarks    : This is Like the CMD Packet Version but will do only one transfer
 // History    : 
 //-----------------------------------------------------------------------------
-int DMARead(void* pci,void* vme, unsigned int count, int ctl)
+void VME_DMA(void* pci, void* vme, unsigned int count, int ctl, TDMAcallback cback)
 {
-  int retval = 0;
-  int val = 0;
-  int timeout = 0;
-  char buf[256];
-
 #ifdef __DEBUG__
-  sprintf(buf,"<<ca91c042>> DMA Read Request of virtpci=0x%08X vme=0x%08X cnt=%i ctl=0x%08X\n",
+  char buf[256];
+  sprintf(buf,"<<ca91c042>> DMA Request of virtpci=0x%08X vme=0x%08X cnt=%i ctl=0x%08X\n",
           pci,vme,count,ctl);
   printk(buf);
 #endif
@@ -1212,58 +1203,8 @@ int DMARead(void* pci,void* vme, unsigned int count, int ctl)
   writel(virt_to_bus(pci), baseaddr + DLA);
   writel((int)vme, baseaddr + DVA);
 
-  // Start the Transfer
-  writel(0x80006F00,baseaddr+DGCS);
-
-  val = readl(baseaddr+DGCS);
-  // THIS IS ONLY A TEST, THIS IS NOT READY FOR MAINSTREAM USE
-  while((val & 0x00008000) && (timeout++ < 100000))
-    val = readl(baseaddr+DGCS);
-
-  if (timeout == 100000)
-    printk("<<ca91c042>> DMA Timed out in DMARead\n");
-
-  if (!(val & 0x00000800)) {
-    sprintf(buf,"<<ca91c042>> DMA Read Error %08X\n",val);  
-    printk(buf);
-    sprintf(buf,"<<ca91c042>> DMA Read Request of virtpci=0x%08X, buspci=%08X vme=0x%08X\n",
-            pci, virt_to_bus(pci), vme);
-    printk(buf);
-    sprintf(buf,"<<ca91c042>> DMA Read Request of count=%i ctl=0x%08X\n",count, ctl);
-    printk(buf);
-    val = readl(baseaddr+DVA);
-    sprintf(buf,"<<ca91c042>> DMA Read Error- from Register vme_addr=0x%08X\n",val);
-    printk(buf);
-    val = readl(baseaddr+DLA);
-    sprintf(buf,"<<ca91c042>> DMA Read Error- from register pci_addr=0x%08X\n",val);
-    printk(buf);
-  }
-
-  retval = readl(baseaddr+DTBC);  // Get Number of bytes transfered
-
-  return retval;
-}
-
-//-----------------------------------------------------------------------------
-// Function   : DMAReadCallBack
-// Inputs     : void* CmdPacketList,TDMAcallback cback
-// Outputs    : int
-// Description: 
-// Remarks    : CmdPacketList is a Phys Address
-// History    : 
-//-----------------------------------------------------------------------------
-void DMAReadCallBack(void* CmdPacketList,TDMAcallback cback)
-{
-  debugptr = CmdPacketList;
-
-  // Setup a DMA Transfer and once the Transfer is complete call cback
-  writel(0x08006F00,baseaddr+DGCS);
-  writel(0x00000000,baseaddr+DTBC);
-  writel((int)CmdPacketList,baseaddr+DCPP);   // Command Packet Pointer
-
   // We need to build a timer to timeout DMA access
-  DMA_timer.expires  = jiffies + 1*HZ;     
-  DMA_timer.function = DMA_status;
+  DMA_timer.expires  = jiffies + DMATIMEOUT;     
   add_timer(&DMA_timer);
 
   // Setup CallBack Function
@@ -1271,8 +1212,60 @@ void DMAReadCallBack(void* CmdPacketList,TDMAcallback cback)
 
   // Start the Transfer,
   // Interrupt on Stop,Halt,Done,LERR,VERR,Prot Error
-  writel(0x88006F6F,baseaddr+DGCS);           
+  DMAType = DMATYPE_SINGLE;
+  writel(0x80216F6F,baseaddr+DGCS);           
 
   // We are now all setup, so lets return to the calling function so it can
   // sleep or what ever until the DMA is done
 }
+
+//-----------------------------------------------------------------------------
+// Function   : VME_DMA_LinkedList
+// Inputs     : void* CmdPacketList,TDMAcallback cback
+// Outputs    : int
+// Description: 
+// Remarks    : CmdPacketList is a Phys Address
+// History    : 
+//-----------------------------------------------------------------------------
+void VME_DMA_LinkedList(void* CmdPacketList,TDMAcallback cback)
+{
+  debugptr = CmdPacketList;
+
+  // Setup a DMA Transfer and once the Transfer is complete call cback
+  writel(0x08316F00,baseaddr+DGCS);
+  writel(0x00000000,baseaddr+DTBC);
+  writel((int)CmdPacketList,baseaddr+DCPP);   // Command Packet Pointer
+
+  // We need to build a timer to timeout DMA access
+  DMA_timer.expires  = jiffies + DMATIMEOUT;     
+  add_timer(&DMA_timer);
+
+  // Setup CallBack Function
+  DMACallBackFunc = cback;
+
+  // Start the Transfer,
+  // Interrupt on Stop,Halt,Done,LERR,VERR,Prot Error
+  DMAType = DMATYPE_LLIST;
+  writel(0x88216F6F,baseaddr+DGCS);           
+
+  // We are now all setup, so lets return to the calling function so it can
+  // sleep or what ever until the DMA is done
+}
+
+//-----------------------------------------------------------------------------
+// Function   : VME_Bus_Error
+// Inputs     : 
+// Outputs    : 
+// Description: Returns 1 if a Bus Error is Pending
+// Remarks    : 
+// History    : 10/30/2000 mjw Started
+//-------------------------------------------------------------------------mjw-
+int VME_Bus_Error(void)
+{
+  int pci_csr;
+
+  pci_csr = readl(baseaddr+PCI_CSR);           // Check for Bus Error
+  writel(pci_csr,baseaddr+PCI_CSR);            // Clear the Error
+  return((pci_csr & 0x08000000) ? 1 : 0);
+}
+
